﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using DiffCommon;
using System.Runtime.Serialization;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Diagnostics;

namespace DiffControls
{
	public class ResultBag
	{
		public char[] A { get; protected set; }
		public char[] B { get; protected set; }

		public Results Forward { get; protected set; }
		public Results Reverse { get; protected set; }
		public Results Linear { get; protected set; }

		public ResultBag( char[] a, char[] b, Results forward, Results reverse, Results linear )
		{
			A = a;
			B = b;

			Forward = forward;
			Reverse = reverse;
			Linear = linear;
		}
	}

	public enum ResultType
	{
		None,
		Forward,
		Reverse,
		Linear,
	}

	//-----------------------------------------------------------------------------------------------

	[Serializable]
	class Options : ISerializable
	{

		//-----------------------------------------------------------------------------------------------

		public bool DrawIdeal { get; set; }
		public bool DrawDiagonals { get; set; }
		public bool DrawSolution { get; set; }
		public bool DrawEvenK { get; set; }
		public bool DrawOddK { get; set; }
		public bool DrawStats { get; set; }

		public ResultType ResultType { get; set; }

		public SnakeOptions SnakeOptions { get; set; }
		public ContourOptions ContourOptions { get; set; }

		public string A { get; set; }
		public string B { get; set; }

		//-----------------------------------------------------------------------------------------------

		public Options() { }

		//-----------------------------------------------------------------------------------------------

		public void Save( string filepath )
		{
			try
			{
				using ( var fs = File.OpenWrite( filepath ) )
				{
					var s = new BinaryFormatter();
					s.Serialize( fs, this );
				}
			}
			catch ( Exception x )
			{
				Trace.WriteLine( "Options.Save( " + filepath + " )\n" + x );
			}
		}

		public static Options Load( string filepath )
		{
			if ( !File.Exists( filepath ) ) return null;

			try
			{
				using ( var fs = File.OpenRead( filepath ) )
				{
					var s = new BinaryFormatter();
					var o = s.Deserialize( fs );
					return ( o as Options );
				}
			}
			catch ( Exception x )
			{
				Trace.WriteLine( "Options.Load( " + filepath + " )\n" + x );
				return null;
			}
		}

		//-----------------------------------------------------------------------------------------------

		// save
		public void GetObjectData( SerializationInfo info, StreamingContext context )
		{
			info.AddValue( "version", 1 );

			info.AddValue( "DrawIdeal", DrawIdeal );
			info.AddValue( "DrawDiagonals", DrawDiagonals );
			info.AddValue( "DrawSolution", DrawSolution );
			info.AddValue( "DrawEvenK", DrawEvenK );
			info.AddValue( "DrawOddK", DrawOddK );
			info.AddValue( "DrawStats", DrawStats );

			info.AddValue( "ResultType", ResultType );

			info.AddValue( "SnakeOptions", SnakeOptions );
			info.AddValue( "ContourOptions", ContourOptions );

			info.AddValue( "A", A );
			info.AddValue( "B", B );
		}

		// load
		Options( SerializationInfo info, StreamingContext context )
		{
			int version = info.GetInt32( "version" );

			DrawIdeal = info.GetBoolean( "DrawIdeal" );
			DrawDiagonals = info.GetBoolean( "DrawDiagonals" );
			DrawSolution = info.GetBoolean( "DrawSolution" );
			DrawEvenK = info.GetBoolean( "DrawEvenK" );
			DrawOddK = info.GetBoolean( "DrawOddK" );
			DrawStats = info.GetBoolean( "DrawStats" );

			ResultType = ( ResultType ) info.GetValue( "ResultType", typeof( ResultType ) );

			SnakeOptions = ( SnakeOptions ) info.GetValue( "SnakeOptions", typeof( SnakeOptions ) );
			ContourOptions = ( ContourOptions ) info.GetValue( "ContourOptions", typeof( ContourOptions ) );

			A = info.GetString( "A" );
			B = info.GetString( "B" );
		}

		//-----------------------------------------------------------------------------------------------

	}

	//-----------------------------------------------------------------------------------------------

	[Serializable]
	class DOptions : ISerializable
	{
		public bool Show { get; protected set; }

		public int Start { get; set; }
		public int Step { get; set; }
		public int End { get; set; }

		protected DOptions() { }

		protected DOptions( bool show, int start, int step, int end )
		{
			Show = show;

			Start = ( start >= 0 ? start : 0 );
			Step = ( step >= 1 ? step : 1 );
			End = ( end >= 0 ? end : 0 );
		}

		// save
		public void GetObjectData( SerializationInfo info, StreamingContext context )
		{
			info.AddValue( "version", 1 );

			info.AddValue( "Show", Show );

			info.AddValue( "Start", Start );
			info.AddValue( "Step", Step );
			info.AddValue( "End", End );
		}

		// load
		protected DOptions( SerializationInfo info, StreamingContext context )
		{
			int version = info.GetInt32( "version" );

			Show = info.GetBoolean( "Show" );

			Start = info.GetInt32( "Start" );
			Step = info.GetInt32( "Step" );
			End = info.GetInt32( "End" );
		}
	}

	[Serializable]
	class SnakeOptions : DOptions
	{
		static SnakeOptions _Off = new SnakeOptions { Show = false };
		public static SnakeOptions Off { get { return _Off; } }

		SnakeOptions() { }

		public SnakeOptions( bool show, int start, int step, int end )
			: base( show, start, step, end )
		{
		}

		// load
		SnakeOptions( SerializationInfo info, StreamingContext context ) : base( info, context ) { }
	}

	[Serializable]
	class ContourOptions : DOptions
	{
		static ContourOptions _Off = new ContourOptions { Show = false };
		public static ContourOptions Off { get { return _Off; } }

		ContourOptions() { }

		public ContourOptions( bool show, int start, int step, int end )
			: base( show, start, step, end )
		{
		}

		// load
		ContourOptions( SerializationInfo info, StreamingContext context ) : base( info, context ) { }
	}

	//-----------------------------------------------------------------------------------------------

}
