﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;

using Common;

using DiffCommon;

namespace DiffControls
{
	public partial class GridUserControl : UserControl
	{
		public GridUserControl()
		{
			InitializeComponent();

			HandleDestroyed += ( s, e ) => SaveOptions();

			scrollZoom.ValueChanged += HandlesZoom;

			MouseWheel += new MouseEventHandler( HandlesMouseWheel );

			rbNone.CheckedChanged += new EventHandler( rbAlgorithm_CheckedChanged );
			rbForward.CheckedChanged += new EventHandler( rbAlgorithm_CheckedChanged );
			rbReverse.CheckedChanged += new EventHandler( rbAlgorithm_CheckedChanged );
			rbLinear.CheckedChanged += new EventHandler( rbAlgorithm_CheckedChanged );

			cbDrawIdeal.CheckedChanged += new EventHandler( cbDrawIdeal_CheckedChanged );
			cbDrawDiagonals.CheckedChanged += new EventHandler( cbDrawDiagonals_CheckedChanged );
			cbDrawSolution.CheckedChanged += new EventHandler( cbDrawSolution_CheckedChanged );
			cbDrawEvenK.CheckedChanged += new EventHandler( cbDrawEvenK_CheckedChanged );
			cbDrawOddK.CheckedChanged += new EventHandler( cbDrawOddK_CheckedChanged );
			cbDrawStats.CheckedChanged += new EventHandler( cbDrawStats_CheckedChanged );

			cbContourShow.CheckedChanged += new EventHandler( cbContourShow_CheckedChanged );
			udContourStart.ValueChanged += new EventHandler( udContourStart_ValueChanged );
			udContourStep.ValueChanged += new EventHandler( udContourStep_ValueChanged );
			udContourEnd.ValueChanged += new EventHandler( udContourEnd_ValueChanged );

			cbSnakeShow.CheckedChanged += new EventHandler( cbSnakeShow_CheckedChanged );
			udSnakeStart.ValueChanged += new EventHandler( udSnakeStart_ValueChanged );
			udSnakeStep.ValueChanged += new EventHandler( udSnakeStep_ValueChanged );
			udSnakeEnd.ValueChanged += new EventHandler( udSnakeEnd_ValueChanged );

			A.TextChanged += AB_TextChanged;
			B.TextChanged += AB_TextChanged;

			new WinFormsKickstart( this ).Loaded += HandlesLoaded;
		}

		void HandlesLoaded()
		{
			LoadOptions();
		}

		//-----------------------------------------------------------------------------------------------

		//public void Init( string a, string b )
		//{
		//    _CalculateInternal = true;
		//    A.Text = a;
		//    B.Text = b;
		//    _CalculateInternal = false;

		//    Calculate( a, b );

		//    LoadOptions();
		//}

		bool _CalculateInternal = false;
		void AB_TextChanged( object sender, EventArgs e )
		{
			lblN.Text = "N: " + A.Text.Length;
			lblM.Text = "M: " + B.Text.Length;

			if ( _CalculateInternal ) return;

			Calculate( A.Text, B.Text );
		}

		//-----------------------------------------------------------------------------------------------

		void Calculate()
		{
			Calculate( A.Text, B.Text );
		}

		void Calculate( string a, string b )
		{
			var ca = a.ToCharArray();
			var cb = b.ToCharArray();

			SetResults(
				new DiffControls.ResultBag(
					ca, cb,
					DiffGreedy.Diff.Compare( ca, cb, true ),
					DiffGreedy.Diff.Compare( ca, cb, false ),
					DiffLinear.Diff.Compare( ca, cb ) ) );
		}

		void SetResults( ResultBag r )
		{
			SetText( r.A, r.B );

			grid.SetResults( r );

			scrollZoom.Value = grid.ZoomSliderFit;
		}

		void SetText( char[] a, char[] b )
		{
			int max = a.Length + b.Length;

			udSnakeStart.Maximum = max;
			udSnakeStep.Maximum = max;
			udSnakeEnd.Maximum = max;

			udContourStart.Maximum = max;
			udContourStep.Maximum = max;
			udContourEnd.Maximum = max;
		}

		//-----------------------------------------------------------------------------------------------

		string OptionsFilepath { get { return Path.Combine( DiffTutorial.Glob.DirUserRoaming, "options.bin" ); } }

		void LoadOptions()
		{
			Debug.WriteLine( "GridUserControl.LoadOptions( " + OptionsFilepath + " )" );

			var o = Options.Load( OptionsFilepath );
			if ( o == null )
			{
				_CalculateInternal = true;
				A.Text = "ABCABBA";
				B.Text = "CBABAC";
				_CalculateInternal = false;

				Calculate();

				return;
			}

			_CalculateInternal = true;
			A.Text = o.A;
			B.Text = o.B;
			_CalculateInternal = false;

			Calculate();

			cbDrawIdeal.Checked = o.DrawIdeal;
			cbDrawDiagonals.Checked = o.DrawDiagonals;
			cbDrawSolution.Checked = o.DrawSolution;
			cbDrawEvenK.Checked = o.DrawEvenK;
			cbDrawOddK.Checked = o.DrawOddK;
			cbDrawStats.Checked = o.DrawStats;

			ResultType = o.ResultType;

			SnakeOptions = o.SnakeOptions;
			ContourOptions = o.ContourOptions;
		}

		void SaveOptions()
		{
			Debug.WriteLine( "GridUserControl.SaveOptions( " + OptionsFilepath + " )" );

			var o = new Options();

			o.DrawIdeal = cbDrawIdeal.Checked;
			o.DrawDiagonals = cbDrawDiagonals.Checked;
			o.DrawSolution = cbDrawSolution.Checked;
			o.DrawEvenK = cbDrawEvenK.Checked;
			o.DrawOddK = cbDrawOddK.Checked;
			o.DrawStats = cbDrawStats.Checked;

			o.ResultType = ResultType;

			o.SnakeOptions = SnakeOptions;
			o.ContourOptions = ContourOptions;

			o.A = A.Text;
			o.B = B.Text;

			o.Save( OptionsFilepath );
		}

		//-----------------------------------------------------------------------------------------------

		void HandlesZoom( object sender, EventArgs e )
		{
			try { grid.Zoom = scrollZoom.Value; }
			catch { }
		}

		void HandlesMouseWheel( object sender, MouseEventArgs e )
		{
			if ( e.Delta == 0 ) return;

			var v = scrollZoom.Value + ( e.Delta > 0 ? 10 : -10 );

			if ( v < 0 ) v = 0;
			if ( v > 1000 ) v = 1000;

			scrollZoom.Value = v;
		}

		//-----------------------------------------------------------------------------------------------
		// options

		void cbDrawIdeal_CheckedChanged( object sender, EventArgs e )
		{
			grid.DrawIdeal = cbDrawIdeal.Checked;
		}

		void cbDrawDiagonals_CheckedChanged( object sender, EventArgs e )
		{
			grid.DrawDiagonals = cbDrawDiagonals.Checked;
		}

		void cbDrawSolution_CheckedChanged( object sender, EventArgs e )
		{
			grid.DrawSolution = cbDrawSolution.Checked;
		}

		void cbDrawEvenK_CheckedChanged( object sender, EventArgs e )
		{
			grid.DrawEvenK = cbDrawEvenK.Checked;
		}

		void cbDrawOddK_CheckedChanged( object sender, EventArgs e )
		{
			grid.DrawOddK = cbDrawOddK.Checked;
		}

		void cbDrawStats_CheckedChanged( object sender, EventArgs e )
		{
			grid.DrawStats = cbDrawStats.Checked;
		}

		//-----------------------------------------------------------------------------------------------
		// Algorithm

		void rbAlgorithm_CheckedChanged( object sender, EventArgs e )
		{
			grid.ResultType = ResultType;
		}

		ResultType ResultType
		{
			get
			{
				if ( rbForward.Checked ) return ResultType.Forward;
				if ( rbReverse.Checked ) return ResultType.Reverse;
				if ( rbLinear.Checked ) return ResultType.Linear;
				return ResultType.None;
			}

			set
			{
				switch ( value )
				{
					case ResultType.Forward: rbForward.Checked = true; break;
					case ResultType.Reverse: rbReverse.Checked = true; break;
					case ResultType.Linear: rbLinear.Checked = true; break;
					default: rbNone.Checked = true; break;
				}
			}
		}

		//-----------------------------------------------------------------------------------------------
		// D Snake

		void cbSnakeShow_CheckedChanged( object sender, EventArgs e )
		{
			SetSnake();
		}

		void udSnakeStart_ValueChanged( object sender, EventArgs e )
		{
			if ( udSnakeStart.Value > udSnakeEnd.Value ) udSnakeEnd.Value = udSnakeStart.Value;

			SetSnake();
		}

		void udSnakeStep_ValueChanged( object sender, EventArgs e )
		{
			SetSnake();
		}

		void udSnakeEnd_ValueChanged( object sender, EventArgs e )
		{
			if ( udSnakeEnd.Value < udSnakeStart.Value ) udSnakeStart.Value = udSnakeEnd.Value;

			SetSnake();
		}

		void SetSnake()
		{
			grid.SnakeOptions = ( !cbSnakeShow.Checked ) ? SnakeOptions.Off : SnakeOptions;
		}

		SnakeOptions SnakeOptions
		{
			get
			{
				return new SnakeOptions(
					cbSnakeShow.Checked,
					( int ) udSnakeStart.Value,
					( int ) udSnakeStep.Value,
					( int ) udSnakeEnd.Value );
			}

			set
			{
				cbSnakeShow.Checked = value.Show;

				SetUpDown( udSnakeStart, value.Start );
				SetUpDown( udSnakeStep, value.Step );
				SetUpDown( udSnakeEnd, value.End );
			}
		}

		//-----------------------------------------------------------------------------------------------
		// D Contour

		void cbContourShow_CheckedChanged( object sender, EventArgs e )
		{
			SetContour();
		}

		void udContourStart_ValueChanged( object sender, EventArgs e )
		{
			if ( udContourStart.Value > udContourEnd.Value ) udContourEnd.Value = udContourStart.Value;

			SetContour();
		}

		void udContourStep_ValueChanged( object sender, EventArgs e )
		{
			SetContour();
		}

		void udContourEnd_ValueChanged( object sender, EventArgs e )
		{
			if ( udContourEnd.Value < udContourStart.Value ) udContourStart.Value = udContourEnd.Value;

			SetContour();
		}

		void SetContour()
		{
			grid.ContourOptions = ( !cbContourShow.Checked ) ? ContourOptions.Off : ContourOptions;

		}

		ContourOptions ContourOptions
		{
			get
			{
				return new ContourOptions(
					cbContourShow.Checked,
					( int ) udContourStart.Value,
					( int ) udContourStep.Value,
					( int ) udContourEnd.Value );
			}

			set
			{
				cbContourShow.Checked = value.Show;

				SetUpDown( udContourStart, value.Start );
				SetUpDown( udContourStep, value.Step );
				SetUpDown( udContourEnd, value.End );
			}
		}

		//-----------------------------------------------------------------------------------------------

		void SetUpDown( NumericUpDown ud, int i )
		{
			if ( i < ud.Minimum ) { ud.Value = ud.Minimum; return; }
			if ( i > ud.Maximum ) { ud.Value = ud.Maximum; return; }
			ud.Value = i;
		}

		//-----------------------------------------------------------------------------------------------

	}
}
